/*
 *      backend-operations.c
 *
 *      Copyright 2007 Florian Sievers <florian@dopehouse.dyndns.org>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <gtk/gtk.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>

#include "support.h"

#include "backend-operations.h"

/* global variables */
mimms_param mimms_parameters;
fork_data child_data;

#define PREFERENCE_DIR_CREATE_MODE  (S_IRUSR | S_IWUSR | S_IXUSR)
#define PREFERENCE_FILE_CREATE_MODE (S_IRUSR | S_IWUSR)

#ifndef BUFSIZ
#define BUFSIZ  8192
#endif

/* shortcuts for gtk_widget_set_sensitive function */
#define SENSITIVE(WIDGET, NAME) gtk_widget_set_sensitive(lookup_widget(WIDGET, NAME), TRUE)
#define INSENSITIVE(WIDGET, NAME) gtk_widget_set_sensitive(lookup_widget(WIDGET, NAME), FALSE)

/* local functions */
static gchar *preference_get_filename(void);
mms_methods mimms_get_method_from_url(gchar *url);
gboolean mimms_replace_method(mms_methods new_mms_method);
gboolean update_download_window_callback(gpointer data);
gint read_data(void);
static int read_from_pipe(int pipefd, char *buf, int bufsz);
char *next_line(char *buf, int bufsz, char *line);


/**
 * name: preference_load
 *
 * Return TRUE if file could be loaded and FALSE on error.
 *
 * Loads the config file and sets the preferences.
 *
 * Returns: a boolean.
 **/
gboolean preference_load()
{
	GKeyFile *keyFile;
	gboolean result;
	gchar *filename = NULL;

	/* get the name of the preferencefile */
	filename = preference_get_filename();

	keyFile = g_key_file_new();

	result = g_key_file_load_from_file(keyFile, filename, G_KEY_FILE_KEEP_COMMENTS, NULL);
	if(result == FALSE)
	{
		/* create new file, if file doesn't exist */
		close(g_creat(filename, PREFERENCE_FILE_CREATE_MODE));

		/* try locating the mimms binary */
		set_mimms_path(find_mimms_binary());

		g_key_file_free(keyFile);
		g_free(filename);

		/* save the preferences */
		preference_save();

		return FALSE;
	}

	/* get the path to mimms binary from the config file */
	set_mimms_path(g_key_file_get_string(keyFile, "general", "mimms_path", NULL));
	if(get_mimms_path == NULL)
	{
		/* try locating the mimms binary, if the config file didn't
		 * have it */
		set_mimms_path(find_mimms_binary());

		g_key_file_free(keyFile);
		g_free(filename);

		return FALSE;
	}

	g_key_file_free(keyFile);
	g_free(filename);
	return TRUE;
}


/**
 * preference_save:
 *
 * Return TRUE if file could be saved and FALSE on error.
 *
 * Saves the preferences to the config file.
 *
 * Returns: a boolean.
 **/
gboolean preference_save()
{
	GKeyFile *keyFile;
	gboolean result;
	gchar *filename = NULL;
	gchar *data = NULL;

	/* get the name of the preferencefile */
	filename = preference_get_filename();

	keyFile = g_key_file_new();

	result = g_key_file_load_from_file(keyFile, filename, G_KEY_FILE_NONE, NULL);
	if (result == FALSE)
	{
		gchar *start_comment = g_strdup_printf(_("%s configuration file, edit as you need"), PACKAGE);
		g_key_file_set_comment(keyFile, NULL, NULL, start_comment, NULL);
		g_free(start_comment);
	}

	/* save the path to the mimms binary to the config file */
	if(get_mimms_path() != NULL)
	{
		g_key_file_set_string(keyFile, "general", "mimms_path", get_mimms_path());
	}

	/* write the configuration data to the config file */
	data = g_key_file_to_data(keyFile, NULL, NULL);
	g_file_set_contents(filename, data, -1, NULL);

	g_key_file_free(keyFile);
	g_free(filename);
	return TRUE;
}


/**
 * preference_get_filename:
 *
 * Returns the filename of the preference file.
 * The returned filename must be freed with g_free() after use.
 *
 * Returns: a static char.
 **/
static gchar *preference_get_filename(void)
{
	gchar *dir;
	gchar *filename;

	/* build directory name (homedir+.PACKAGENAME) */
	dir = g_build_filename (g_get_home_dir (), "."PACKAGE, NULL);
	/* check if directory exists, if not create it */
	if (!g_file_test (dir, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
	{
		g_mkdir_with_parents (dir, PREFERENCE_DIR_CREATE_MODE);
	}

	/* build filename */
	filename = g_build_filename (dir, "preference", NULL);
	g_free(dir);

	return filename;
}


/**
 * find_mimms_binary:
 *
 * Returns the path to mimms or NULL on error.
 * The returned string must be freed with g_free() after use.
 *
 * Returns: a char.
 **/
gchar *find_mimms_binary()
{
	gchar *mimms_executable;
	GtkWidget *dialog;

	mimms_executable = g_find_program_in_path("mimms");

	if(mimms_executable == NULL)
	{
		dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
										GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
										g_locale_to_utf8(_("Could not find MiMMS binary in path.\nPlease specify the path manual in the preference dialog."), -1, NULL, NULL, NULL));

		gtk_dialog_run(GTK_DIALOG (dialog));
		gtk_widget_destroy(dialog);

		return NULL;
	}

	return mimms_executable;
}


/**
 * return_mimms_path:
 *
 * Returns the actual path to the mimms binary.
 * Do not free this string, because it is managed by the backend
 * functions.
 *
 * Returns: a char.
 **/
gchar *get_mimms_path()
{
	return mimms_parameters.path_to_mimms;
}


/**
 * set_mimms_path:
 * @mimms_path: new path to mimms binary
 *
 * Returns the actual path to the mimms binary.
 * Do not free this string, because it is managed by the backend
 * functions.
 **/
void set_mimms_path(gchar *mimms_path)
{
	/* check if path_to_mimms is not NULL and free the memory */
	if(mimms_parameters.path_to_mimms != NULL)
	{
		g_free(mimms_parameters.path_to_mimms);
	}

	/* set net path to mimms */
	mimms_parameters.path_to_mimms = mimms_path;
}


/**
 * backend_init:
 *
 * Initiates the global mimms_parameters structure with empty data.
 **/
void backend_init()
{
	mimms_parameters.use_timer = 0;
	mimms_parameters.timer = 0;
	mimms_parameters.url = NULL;
	mimms_parameters.output_file = NULL;
	mimms_parameters.method = mms_default;
	mimms_parameters.path_to_mimms = NULL;
	preference_load();

	child_data.pipes[0] = -1;
	child_data.pipes[1] = -1;
}


/**
 * backend_cleanup:
 *
 * Deallocates some variables.
 **/
void backend_cleanup()
{
	if(mimms_parameters.path_to_mimms != NULL)
	{
		g_free(mimms_parameters.path_to_mimms);
	}
}


/**
 * mimms_set_method:
 * @mms_methods: mms streaming method
 * @widget: needed to lookup URL entry widget
 * @update_entry_url: set to TRUE if entry_url should be updated or
 * 					  FALSE if not
 *
 * Sets the mms method to be used.
 **/
void mimms_set_method(mms_methods method, GtkWidget *widget, gboolean update_entry_url)
{
	GtkWidget *entry_url;
	gboolean url_has_changed = FALSE;
	static gboolean entry_written = FALSE;

	/* replace old URL prefix with new one */
	if(!mimms_replace_method(method))
	{
		return;
	}

	/* if mimms_parameters.url isn't NULL and update_entry_url is TRUE,
	 * update the entry_url */
	if(mimms_parameters.url != NULL && update_entry_url == TRUE)
	{
		gchar *converted_string;

		converted_string = g_locale_to_utf8(mimms_parameters.url, -1, NULL, NULL, NULL);

		/* udpate the URL entry */
		entry_url = lookup_widget(widget, "entry_url");
		gtk_entry_set_text(GTK_ENTRY(entry_url), converted_string);

		g_free(converted_string);
	}
	return;
}


/**
 * mimms_use_timer:
 * @status: use timer yes(1)/no(0)
 *
 * Enables (1) or disables (0) the timer.
 **/
void mimms_use_timer(int status)
{
	if(status == 0)
	{
		mimms_parameters.use_timer = 0;
	}
	else
	{
		mimms_parameters.use_timer = 1;
	}
}

/**
 * mimms_set_timer:
 * @timer: stream recording time in minutes
 *
 * Sets the recording timer for mimms. The value defines how long mimms
 * should record the stream in minutes.
 **/
void mimms_set_timer(int timer)
{
	mimms_parameters.timer = timer;
}


/**
 * replace_method:
 * @new_mms_method: the new MMS-Method for the URL
 *
 * Returns: TRUE if MMS-Method has been changed, FALSE if not changed
 * Replaces the MMS-Method in the URL.
 **/
gboolean mimms_replace_method(mms_methods new_mms_method)
{
	gchar *temp_url_no_method = NULL;
	gchar *temp_new_url = NULL;
	gchar *new_method_prefix = NULL;

	/* check if new_method == old_method */
	if(new_mms_method == mimms_parameters.method)
	{
		return FALSE;
	}

	/* set new method */
	mimms_parameters.method = new_mms_method;

	/* return if URL is NULL */
	if(mimms_parameters.url == NULL)
	{
		mimms_parameters.url = g_strdup("");
	}

	/* connect new prefix with URL */
	switch(new_mms_method)
	{
		case mms_default:
			new_method_prefix = g_strdup("mms");
			break;

		case mms_tcp:
			new_method_prefix = g_strdup("mmst");
			break;

		case mms_udp:
			new_method_prefix = g_strdup("mmsu");
			break;

		case mms_http:
			new_method_prefix = g_strdup("mmsh");
			break;

		case mms_asx_http:
			new_method_prefix = g_strdup("http");
			break;

		default:
			printf("Error: unknown method\n");
			return FALSE;
			break;
	}

	temp_url_no_method = g_strrstr(mimms_parameters.url, "://");

	if(temp_url_no_method != NULL)
	{
		temp_new_url = g_strconcat(new_method_prefix, temp_url_no_method, NULL);
	}
	else
	{
		if(strlen(mimms_parameters.url) > 0)
		{
			temp_new_url = g_strconcat(new_method_prefix, "://", mimms_parameters.url, NULL);
		}
		else
		{
			temp_new_url = g_strconcat(new_method_prefix, "://", NULL);
		}
	}

	if(mimms_parameters.url != NULL)
	{
		g_free(mimms_parameters.url);
	}
	mimms_parameters.url = g_strdup(temp_new_url);

	g_free(temp_new_url);
	g_free(new_method_prefix);

	return TRUE;
}


/**
 * get_method_from_url:
 * @url: stream url
 *
 * Returns: MMS-Method from url.
 **/
mms_methods mimms_get_method_from_url(gchar *url)
{
	/* check if URL is not NULL or ZERO-length */
	if(url == NULL || strlen(url) == 0)
	{
		return mms_not_defined;
	}

	if(g_str_has_prefix(url, "mms://"))
	{
		return mms_default;
	}
	else if(g_str_has_prefix(url, "mmst://"))
	{
		return mms_tcp;
	}
	/* unsupported mms via udp
	else if(g_str_has_prefix(url, "mmsu://"))
	{
		return mms_udp;
	}
	*/
	else if(g_str_has_prefix(url, "mmsh://"))
	{
		return mms_http;
	}
	else if(g_str_has_prefix(url, "http://"))
	{
		return mms_asx_http;
	}
	else
	{
		return mms_not_defined;
	}
}


/**
 * mimms_set_url:
 * @url: stream url
 *
 * Sets the input URL for mimms.
 **/
void mimms_set_url(gchar *url, GtkWidget *widget)
{
	GtkWidget *radiobutton = NULL;
	gchar *lower_case_url = NULL;

	if(url == NULL)
	{
		fprintf(stderr, "URL is set to null!\n");
		return;
	}

	if(mimms_parameters.url != NULL)
	{
		g_free(mimms_parameters.url);
	}

	mimms_parameters.url = g_strdup(url);

	lower_case_url = g_utf8_strdown(mimms_parameters.url, -1);
	if(lower_case_url == NULL)
	{
		fprintf(stderr, "Error in string allocation!\n");
		return;
	}

	if(g_pattern_match_simple("*://*", lower_case_url) == TRUE)
	{
		/* set the active radiobutton to the URI */
		switch(mimms_get_method_from_url(lower_case_url))
		{
			case mms_default:
				radiobutton = lookup_widget(widget, "radiobutton_mms");
				break;
			case mms_tcp:
				radiobutton = lookup_widget(widget, "radiobutton_mmst");
				break;
			case mms_udp:
				radiobutton = lookup_widget(widget, "radiobutton_mmsu");
				break;
			case mms_http:
				radiobutton = lookup_widget(widget, "radiobutton_mmsh");
				break;
			case mms_asx_http:
				radiobutton = lookup_widget(widget, "radiobutton_http");
				break;
			default:
				radiobutton = NULL;
				break;
		}
		if(radiobutton != NULL)
		{
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radiobutton), TRUE);
		}
		// url matches *://*
	}
	else
	{
		// url doesn't match *://*
	}

	g_free(lower_case_url);

	return;
}


/**
 * mimms_set_output_file:
 * @filename: output filename
 *
 * Sets the output filename for mimms.
 **/
void mimms_set_output_file(char *filename)
{
	mimms_parameters.output_file = filename;
}


/**
 * start_download:
 *
 * Returns: TRUE if download starts; FALSE on error.
 **/
gboolean start_download(GtkWidget *widget)
{
	#define SIZE 32
	char *url = NULL;
	char *output_file = NULL;
	char *timer = NULL;
	GtkWidget *error_dialog;

	/* Check if URL has been set */
	if(mimms_parameters.url == NULL || strlen(mimms_parameters.url) == 0)
	{
		error_dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
										GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
										g_locale_to_utf8(_("Please enter an URL for download."), -1, NULL, NULL, NULL));

		gtk_dialog_run(GTK_DIALOG (error_dialog));
		gtk_widget_destroy(error_dialog);
		return;
	}

	/* check if output file has been set */
	if(mimms_parameters.output_file == NULL || strlen(mimms_parameters.output_file) == 0)
	{
		error_dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
										GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
										g_locale_to_utf8(_("Please enter an output file."), -1, NULL, NULL, NULL));

		gtk_dialog_run(GTK_DIALOG (error_dialog));
		gtk_widget_destroy(error_dialog);
		return;
	}

	/* check if method is set in the url. return FALSE if not */
	if(mimms_get_method_from_url(mimms_parameters.url) == mms_not_defined)
	{
		error_dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
										GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
										g_locale_to_utf8(_("Select a protocol method."), -1, NULL, NULL, NULL));

		gtk_dialog_run(GTK_DIALOG (error_dialog));
		gtk_widget_destroy(error_dialog);

		printf("method not set\n");
		return FALSE;
	}

	/* return FALSE if output_file is "-" */
	if(g_ascii_strncasecmp(g_strstrip(mimms_parameters.output_file), "-", 1) == 0)
	{
		error_dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
										GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE,
										g_locale_to_utf8(_("Please choose another output filename.\n\nThe filename \"-\" is reserved for output to stdout, which is not supported by GMiMMS."), -1, NULL, NULL, NULL));

		gtk_dialog_run(GTK_DIALOG (error_dialog));
		gtk_widget_destroy(error_dialog);

		printf("stdout\n");
		return FALSE;
	}

	/* check if outputfile already exist and ask to overwrite */
	if(g_file_test(mimms_parameters.output_file, G_FILE_TEST_EXISTS))
	{
		error_dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
										GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO,
										g_locale_to_utf8(_("File already exist. Are you sure to overwrite it?"), -1, NULL, NULL, NULL));

		switch(gtk_dialog_run(GTK_DIALOG (error_dialog)))
		{
			case GTK_RESPONSE_YES:
				gtk_widget_destroy(error_dialog);
				break;
			case GTK_RESPONSE_NO:
			default:
				gtk_widget_destroy(error_dialog);
				return FALSE;
				break;
		}
	}

	/* Allocate memory for timer string */
	if((timer = g_malloc0(SIZE)) == NULL)
	{
		fprintf(stderr, "Can't allocate memory!\n");
		return 0;
	}

	/* Set timer string */
	if(mimms_parameters.use_timer == 1)
	{
		snprintf(timer, SIZE, "-t %i", mimms_parameters.timer);
	}
	else
	{
		snprintf(timer, SIZE, "");
	}

	printf("%s %s \"%s\" \"%s\"\n", mimms_parameters.path_to_mimms, timer, mimms_parameters.url, mimms_parameters.output_file);

	/* create pipes. pipes[0] = read; pipes[1] = write*/
	if(pipe(child_data.pipes) == -1)
	{
		/* error message */
		fprintf(stderr, "Can't create pipes!\nerrno=%d (%s)\n", errno, strerror(errno));

		/* Free allocated memory */
		g_free(timer);

		return FALSE;
	}

	/* make a fork */
	child_data.child_pid = fork();

	/* check for errors */
	if(child_data.child_pid == -1)
	{
		/* error message */
		fprintf(stderr, "Fork failed!\nerrno=%d (%s)\n", errno, strerror(errno));

		/* close read-pipe */
		close(child_data.pipes[0]);
		child_data.pipes[0] = -1;
	}
	/* parent */
	if(child_data.child_pid != 0)
	{
		child_data.main_window = widget;

		INSENSITIVE(child_data.main_window, "main_window");

		/* create download status window and show it */
		child_data.download_status_window = (GtkWidget *) create_download_status_window();
		gtk_widget_show(child_data.download_status_window);

		INSENSITIVE(child_data.download_status_window, "button_close_download_window");

		/* create timeout */
		g_timeout_add(125, update_download_window_callback, widget);
	}

	/* child */
	if(child_data.child_pid == 0)
	{
		sleep(5);

		dup2(child_data.pipes[1], 1);
		dup2(child_data.pipes[1], 2);
		fcntl(child_data.pipes[0], F_SETFL, O_NDELAY);

		/* execute mimms with params */
		execlp(mimms_parameters.path_to_mimms, timer, "-c", mimms_parameters.url, mimms_parameters.output_file, NULL);
		exit(0);
	}

	/* close write-pipe */
	close(child_data.pipes[1]);
	child_data.pipes[1] = -1;

	/* Free allocated memory */
	g_free(timer);

	return TRUE;
}


void cancel_download()
{
	if(child_data.child_pid > 0)
	{
		kill(child_data.child_pid, SIGKILL);
	}
}


/**
 *
 * update_download_window_callback:
 * @data
 *
 * Return FALSE to end timeout callback
 **/
gboolean update_download_window_callback(gpointer data)
{
	//GtkWidget *widget = (GtkWidget *) data;

	/* wait for child (none blocking) and change buttons states */
	if((waitpid(0, NULL, WNOHANG) == child_data.child_pid))
	{
		SENSITIVE(child_data.main_window, "main_window");
		SENSITIVE(child_data.download_status_window, "button_close_download_window");
		INSENSITIVE(child_data.download_status_window, "button_cancel_download");

		child_data.child_pid = 0;
		/* close read-pipe */
		if(child_data.pipes[0] != -1)
		{
			close(child_data.pipes[0]);
			child_data.pipes[0] = -1;
		}

		/* quit timeout */
		return FALSE;
	}

	read_data();

	/* puls the progress bar */
	gtk_progress_bar_pulse(GTK_PROGRESS_BAR(lookup_widget(child_data.download_status_window, "progressbar_working_status")));

	return TRUE;
}


gint read_data(void)
{
	static char buf[BUFSIZ+1] = "";
	static int buflen = 0;
	char line[BUFSIZ+1];
	int count;

	GtkWidget *label_mimms_output;
	GtkWidget *progressbar_working_status;

	label_mimms_output = lookup_widget(child_data.download_status_window, "label_mimms_output");
	progressbar_working_status = lookup_widget(child_data.download_status_window, "progressbar_working_status");

	//while((count = read_from_pipe(child_data.pipes[0], buf+buflen, sizeof(buf) - buflen - 1 )) > 0)
	while((count = read_from_pipe(child_data.pipes[0], buf, sizeof(buf) - buflen - 1 )) > 0)
	{
		char *str;

		/* fprintf(stderr, "\nCount was %d\n", count); */
		buflen = count;
		buf[buflen] = '\0';

		for(str = next_line(buf, sizeof(buf) / sizeof(*buf), line); (str != buf) && (str != NULL); str = next_line(buf, sizeof(buf) / sizeof(*buf), line))
		{
			buflen = strlen(buf);
//			print_line(opt.buffer, str);
			//gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progressbar_working_status), line);
			gtk_label_set_text(GTK_LABEL(label_mimms_output), str);
		}
	}

	/*  fprintf(stderr, "Below loop: Count was %d\n", count); */
	if(buflen > 0)
	{
		char *str;

		while((str = next_line(buf, sizeof(buf) / sizeof(*buf), line)) != NULL)
		{
			buflen = strlen(buf);
//			print_line(opt.buffer, str);
			//gtk_progress_bar_set_text(GTK_PROGRESS_BAR(progressbar_working_status), line);
			gtk_label_set_text(GTK_LABEL(label_mimms_output), str);
			if(str == buf)
				break;
		}
	}

	if(!child_data.child_pid || (waitpid(0, NULL, WNOHANG) == child_data.child_pid))
	{
		/* fprintf(stderr, "Program gone, dead, kablooey!\n"); */
		child_data.child_pid = 0;
		if(child_data.pipes[0] != -1)
		{
			close(child_data.pipes[0]);
			child_data.pipes[0] = -1;
		}
		buflen = 0;
		buf[buflen] = '\0';
		//gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(opt.scanButton), 0);
		return 0;
	}
	/* waitpid unix/windoze selector */

	return(1);
}


/* NOTE:  pipefd must be in O_NONBLOCK mode ( via fcntl ) */
static int read_from_pipe(int pipefd, char *buf, int bufsz)
{
	int count;

	if(pipefd == -1)
		return -1;
	count = read(pipefd, buf, bufsz);
	if(count == -1 && errno != EINTR && errno != EAGAIN)
	{
//		pfatal("Failed to read from nmap process pipe");
	}
	return count;
}


/* split buf into first line and remainder by
   copying the first line into line and stripping it from str;
   return the first line from str or NULL if str contains no full line.
   bufsz is the number of chars in buf.
 */
char *next_line(char *buf, int bufsz, char *line)
{
	if((buf != NULL) && (line != NULL))
	{
		char *eol = strchr(buf, '\n');

		if(eol != NULL)
		{
			char *bol = buf;
			int linelen = MIN(bufsz - 1, eol - buf + 1); // we can't exceed buffer size

			/* copy line including \n to line */
			memcpy(line, buf, linelen);
			line[linelen] = '\0';

			eol = buf + linelen;

			/* remove line from str */
			while(*eol != '\0')
			{
				*bol++ = *eol++;
				if(*bol == '\r')
					*bol = ' ';
				if(*bol == '\n')
					*bol = ' ';
			}
			*bol = '\0';

			return(line);
		}
		return(buf);
	}
	return(NULL);
}
